
//#define INNER_RADIUS                      // Planetary inner radius (earth radius)
#define INTEGRAL_EQUATION_NUM_SAMPLES 2     // Number of sample rays to use in integral equation

// Used by atmospheric scattering
//float4  g_vCameraPos_Height;              // .xyz = camera position, .w = camera height (from center of the earth)
//float3  g_vInvWavelength;                 // 1 / pow(wavelength, 4) for the red, green, and blue channels
//float4  g_vKrKm;                          // .x = Kr * ESun, .y = Km * ESun, .z = Kr * 4 * PI, .w = Km * 4 * PI
//float3  g_vScale;                         // .x = 1 / (fOuterRadius - fInnerRadius), .y =  The scale depth (i.e. the altitude at which the atmosphere's average density is found), .z =  g_fScale / g_fScaleDepth
//float2  g_fMiePhaseAsymmetryFactor;       // x. = Mie phase asymetry factor, .y = .x * .x

float3 GetSunDirection()
{
  return g_vSunDirection;
}

//-----------------------------------------------------------------------------
// Functions used by atmospheric scatteting
//-----------------------------------------------------------------------------
float ExpScale(float fCos)
{
  float x = 1.0 - fCos;
  return g_vScale.y * exp(-0.00287 + x*(0.459 + x*(3.83 + x*(-6.80 + x*5.25))));
}

float GetRayleighPhase(float fCos2)
{
  return 0.75 * (1.0 + fCos2);
}

float GetMiePhase(float fCos, float fCos2)
{
  const float g = g_fMiePhaseAsymmetryFactor.x;
  const float g2 = g_fMiePhaseAsymmetryFactor.y;
  return 1.5 * ((1.0 - g2) / (2.0 + g2)) * (1.0 + fCos2) / pow(max(1.0 + g2 - 2.0*g*fCos, 0.0), 1.5);
}

//-----------------------------------------------------------------------------
// Compute atmospheric scattering per vertex data
//-----------------------------------------------------------------------------
void ComputeAtmosphericScatteringPerVertexData(in float3 i_vPositionEarthCenter, out float3 o_vRayleighColor, out float3 o_vMieColor)
{
  const float fKrESun = g_vKrKm.x;
  const float fKmESun = g_vKrKm.y;
  const float fKr4PI = g_vKrKm.z;
  const float fKm4PI = g_vKrKm.w;

  // Get the ray from the camera to the vertex, and its length (which is the far point of the ray passing through the atmosphere)
  float3 vCameraPosition = g_vCameraPos_Height.xyz;
  float fCameraHeight = g_vCameraPos_Height.w;
  float3 vRay = i_vPositionEarthCenter - vCameraPosition;
  float fFar = length(vRay);
  vRay /= fFar;

  // Calculate the ray's starting position, then calculate its scattering offset
  float3 vStart = vCameraPosition;
  float fHeight = length(vStart);
  float fDepth = exp(g_vScale.z * (INNER_RADIUS - fCameraHeight));
  float fStartAngle = dot(vRay, vStart) / fHeight;
  float fStartOffset = fDepth*ExpScale(fStartAngle);

  // Initialize the scattering loop variables
  float fSampleLength = fFar / float(INTEGRAL_EQUATION_NUM_SAMPLES);
  float fScaledLength = fSampleLength * g_vScale.x;
  float3 vSampleRay = vRay * fSampleLength;
  float3 vSamplePoint = vStart + vSampleRay * 0.5;

  // Now loop through the sample rays
  float3 vFrontColor = float3(0.0, 0.0, 0.0);
  for (int i = 0; i < INTEGRAL_EQUATION_NUM_SAMPLES; i++)
  {
    float fHeight = length(vSamplePoint);
    float fDepth = exp(g_vScale.z * (INNER_RADIUS - fHeight));
    float fLightAngle = dot(GetSunDirection(), vSamplePoint) / fHeight;
    float fCameraAngle = dot(vRay, vSamplePoint) / fHeight;
    float fScatter = (fStartOffset + fDepth*(ExpScale(fLightAngle) - ExpScale(fCameraAngle)));
    float3 vAttenuate = exp(-fScatter * (g_vInvWavelength * fKr4PI + fKm4PI));
    vFrontColor += vAttenuate * (fDepth * fScaledLength);
    vSamplePoint += vSampleRay;
  }

  // Finally, scale the Mie and Rayleigh colors and set up the varying variables for the pixel shader
  o_vMieColor = vFrontColor * fKmESun;
  o_vRayleighColor = vFrontColor * (g_vInvWavelength * fKrESun);
}

//-----------------------------------------------------------------------------
// Compute atmospheric scattering color
//-----------------------------------------------------------------------------
float3 ComputeAtmosphericScatteringColor(in float3 i_vRayleighColor, in float3 i_vMieColor, in float i_fCos, in float i_fSkydomeSunnySide)
{
  const float fCos2 = i_fCos * i_fCos;

  // Compute Rayleigh color
  float fRayleighPhase = GetRayleighPhase(fCos2 * i_fSkydomeSunnySide);
  float3 vRayleighColor = fRayleighPhase * i_vRayleighColor;
  float3 vRayleighColorToneMapped = pow(max(vRayleighColor, 0.0), 1.0 / 2.5);
  vRayleighColor = lerp(vRayleighColor, vRayleighColorToneMapped, max(GetSunDirection().y, 0.0));
  vRayleighColor = saturate(vRayleighColor);

  // Compute Mie color
  float fMiePhase = GetMiePhase(i_fCos, fCos2);
  float3 vMieColor = fMiePhase * i_vMieColor;

  // Return Rayleigh with Mie
  return vRayleighColor + vMieColor;
}
