#include "Noise.es30.glsl"

vec3 decodeDisplacement(vec3 value)
{
#ifdef COGS_BASIC_OCEAN_ENCODING_SCALED
  return valueScale.x * (value - vec3(0.5));
#else
  return value;
#endif
}

vec4 decodeDifference(vec4 value)
{
#ifdef COGS_BASIC_OCEAN_ENCODING_SCALED
  return vec4(valueScale.y * (value.xy - vec2(0.5)),
                1.0 - valueScale.y * value.z,
                0.0);
#else
  return value;
#endif
}

void baseTileCoords(out ivec2 tile, out vec2 pos, in vec2 world)
{
  vec4 tileOffset = vec4(0, 0, 0, 0);     // Tile offset from origin,
  vec4 tileRemainder = vec4(0, 0, 0, 0);  // TIle remainder from origin.

  vec2 t = fftTileScale * world;// +tileRemainder.xy;
  tile = /*tileOffset.xy + */ ivec2(floor(t));
  pos = t - floor(t);
}


vec4 tileSamplePosition(float rnd0, float rnd1, float rnd2, float rnd3, in vec2 pos)
{
  float swellDirection = 0.0;                 //< Direction of waves, zero implies waves that arrive from positive X.
  float oneOverTileExtentAdjust = 1.0;        // Tile extent adjust for when dominant wavelength and tilextent doesn't match.

  float scale = oneOverTileExtentAdjust * (1.4 - (0.8 / 287.0)*rnd3);
  vec2 shifted = scale * (pos + (1.3 / 287.0)*vec2(rnd0, rnd1));
  float twist = (0.3 / 287.0)*(rnd2 - 143.0) + swellDirection;
  vec2 r = vec2(cos(twist), sin(twist));

  return vec4(r.x*shifted.x - r.y*shifted.y,
                r.y*shifted.x + r.x*shifted.y,
                r.x,
                r.y);
}

void setupTileBlend(out vec4 sample00,
                    out vec4 sample01,
                    out vec4 sample10,
                    out vec4 sample11,
                    in ivec2 tile,
                    in vec2 pos)
{
  ivec2 tile00 = tile + ivec2(0, 0);
  ivec2 tile01 = tile + ivec2(0, 1);
  ivec2 tile10 = tile + ivec2(1, 0);
  ivec2 tile11 = tile + ivec2(1, 1);

  vec4 r0 = permute(permute(vec4(tile00.x, tile01.x, tile10.x, tile11.x)) + vec4(tile00.y, tile01.y, tile10.y, tile11.y));
  vec4 r1 = permute(r0 + vec4(1, 1, 1, 1));
  vec4 r2 = permute(r1 + vec4(1, 1, 1, 1));
  vec4 r3 = permute(r2 + vec4(1, 1, 1, 1));

  sample00 = tileSamplePosition(r0.x, r1.x, r2.x, r3.x, pos - vec2(0, 0));
  sample01 = tileSamplePosition(r0.y, r1.y, r2.y, r3.y, pos - vec2(0, 1));
  sample10 = tileSamplePosition(r0.z, r1.z, r2.z, r3.z, pos - vec2(1, 0));
  sample11 = tileSamplePosition(r0.w, r1.w, r2.w, r3.w, pos - vec2(1, 1));
}

vec2 inverseRotateXY(in vec2 v, in vec2 cossin)
{
  return vec2(+cossin.x*v.x + cossin.y*v.y,
              -cossin.y*v.x + cossin.x*v.y);
}

vec3 inverseRotateXY(in vec3 v, in vec2 cossin)
{
  return vec3(+cossin.x*v.x + cossin.y*v.y,
              -cossin.y*v.x + cossin.x*v.y,
               v.z);
}

vec3 waveDisplace(in vec2 world, in float lod)
{
  if (lod < lodDisplace.x) { return vec3(0, 0, 0); }

  float lodBlend = 1.0;// saturate(lodDisplace.y*(lod - lodDisplace.x));

  ivec2 baseTile;
  vec2 basePos;
  baseTileCoords(baseTile, basePos, world);

  vec4 p00, p01, p10, p11;
  setupTileBlend(p00, p01, p10, p11, baseTile, basePos);

  vec3 s00 = decodeDisplacement(texture(Displacement, p00.xy, 0.0).rgb);
  vec3 s01 = decodeDisplacement(texture(Displacement, p01.xy, 0.0).rgb);
  vec3 s10 = decodeDisplacement(texture(Displacement, p10.xy, 0.0).rgb);
  vec3 s11 = decodeDisplacement(texture(Displacement, p11.xy, 0.0).rgb);

  // Bilinear weights
  float b00 = (1.0 - basePos.x)*(1.0 - basePos.y);
  float b01 = (1.0 - basePos.x)*basePos.y;
  float b10 = basePos.x*(1.0 - basePos.y);
  float b11 = basePos.x*basePos.y;;


  vec3 s = (b00 * inverseRotateXY(s00, p00.zw) +
              b01 * inverseRotateXY(s01, p01.zw) +
              b10 * inverseRotateXY(s10, p10.zw) +
              b11 * inverseRotateXY(s11, p11.zw));

  return lodBlend * significantWaveHeight * s;
}

vec3 waveNormal(out float foam, in vec2 world, in float lod, in float foamFactor)
{
  if (lod < lodNormal.x) {
    foam = 0.0;
    return vec3(0, 0, 1);
  }
  float lodNormalBlend = clamp(lodNormal.y*(lod - lodNormal.x), 0.0, 1.0);

  ivec2 baseTile;
  vec2 basePos;
  baseTileCoords(baseTile, basePos, world);

  vec4 p00, p01, p10, p11;
  setupTileBlend(p00, p01, p10, p11, baseTile, basePos);
  // Swap tiles such that a tile is always sampled from the same sampler,
  // otherwise the hardware LOD-calculation fails at seams.
  if ((baseTile.x & 1) != 0) {
    vec4 t;
    t = p00; p00 = p10; p10 = t;
    t = p01; p01 = p11; p11 = t;
    basePos.x = 1.f - basePos.x;
  }
  if ((baseTile.y & 1) != 0) {
    vec4 t;
    t = p00; p00 = p01; p01 = t;
    t = p10; p10 = p11; p11 = t;
    basePos.y = 1.f - basePos.y;
  }

  vec4 n00 = decodeDifference(texture(Difference, p00.xy));
  vec4 n01 = decodeDifference(texture(Difference, p01.xy));
  vec4 n10 = decodeDifference(texture(Difference, p10.xy));
  vec4 n11 = decodeDifference(texture(Difference, p11.xy));
  float b00 = (1.0 - basePos.x)*(1.0 - basePos.y);
  float b01 = (1.0 - basePos.x)*basePos.y;
  float b10 = basePos.x*(1.0 - basePos.y);
  float b11 = basePos.x*basePos.y;;

  vec3 n =
    inverseRotateXY(n00.xyz, p00.zw)*b00 +
    inverseRotateXY(n01.xyz, p01.zw)*b01 +
    inverseRotateXY(n10.xyz, p10.zw)*b10 +
    inverseRotateXY(n11.xyz, p11.zw)*b11;

  foam = 
    n00.w * b00 +
    n01.w * b01 +
    n10.w * b10 +
    n11.w * b11;

  float detailBlend = 0.0;
  if (lodDetail.x < lod) {
    detailBlend = clamp(lodDetail.y*(lod - lodDetail.x), 0.0, 1.0);

    vec4 nm = decodeDifference(texture(Difference, 3.3*fftTileScale*world.xy));
    n += 0.5*detailBlend*nm.xyz;
    foam += 0.5*nm.w;

    vec4 nd = decodeDifference(texture(Difference, 17.1*fftTileScale*world.xy));
    n += 0.25*detailBlend*nd.xyz;

    foam += 0.25*nd.w;
  }
  n = normalize(n);

  foam = 0.0;

  return normalize(mix(vec3(0, 0, 1), n, lodNormalBlend));
}
