#include "Noise.hlsl"

float3 decodeDisplacement(float3 value)
{
#ifdef COGS_BASIC_OCEAN_ENCODING_SCALED
  return valueScale.x * (value - (float3)0.5);
#else
  return value;
#endif
}

float4 decodeDifference(float4 value)
{
#ifdef COGS_BASIC_OCEAN_ENCODING_SCALED
  return float4(valueScale.y * (value.xy - (float2)0.5),
                1.0 - valueScale.y * value.z,
                0.0);
#else
  return value;
#endif
}

void baseTileCoords(out int2 tile, out float2 pos, in float2 world)
{
  float4 tileOffset = float4(0, 0, 0, 0);     // Tile offset from origin,
  float4 tileRemainder = float4(0, 0, 0, 0);  // TIle remainder from origin.

  float2 t = fftTileScale * world;// +tileRemainder.xy;
  tile = /*tileOffset.xy + */ floor(t);
  pos = t - floor(t);
}


float4 tileSamplePosition(float rnd0, float rnd1, float rnd2, float rnd3, in float2 pos)
{
  float swellDirection = 0.f;                 //< Direction of waves, zero implies waves that arrive from positive X.
  float oneOverTileExtentAdjust = 1.0;        // Tile extent adjust for when dominant wavelength and tilextent doesn't match.

  float scale = oneOverTileExtentAdjust * (1.4 - (0.8 / 287.0)*rnd3);
  float2 shifted = scale * (pos + (1.3 / 287.0)*float2(rnd0, rnd1));
  float twist = (0.3 / 287.0)*(rnd2 - 143.0) + swellDirection;
  float2 r = float2(cos(twist), sin(twist));

  return float4(r.x*shifted.x - r.y*shifted.y,
                r.y*shifted.x + r.x*shifted.y,
                r.x,
                r.y);
}

void setupTileBlend(out float4 sample00,
                    out float4 sample01,
                    out float4 sample10,
                    out float4 sample11,
                    in int2 tile,
                    in float2 pos)
{
  int2 tile00 = tile + int2(0, 0);
  int2 tile01 = tile + int2(0, 1);
  int2 tile10 = tile + int2(1, 0);
  int2 tile11 = tile + int2(1, 1);

  float4 r0 = permute(permute(float4(tile00.x, tile01.x, tile10.x, tile11.x)) + float4(tile00.y, tile01.y, tile10.y, tile11.y));
  float4 r1 = permute(r0 + float4(1, 1, 1, 1));
  float4 r2 = permute(r1 + float4(1, 1, 1, 1));
  float4 r3 = permute(r2 + float4(1, 1, 1, 1));

  sample00 = tileSamplePosition(r0.x, r1.x, r2.x, r3.x, pos - int2(0, 0));
  sample01 = tileSamplePosition(r0.y, r1.y, r2.y, r3.y, pos - int2(0, 1));
  sample10 = tileSamplePosition(r0.z, r1.z, r2.z, r3.z, pos - int2(1, 0));
  sample11 = tileSamplePosition(r0.w, r1.w, r2.w, r3.w, pos - int2(1, 1));
}

float2 inverseRotateXY(in float2 v, in float2 cossin)
{
  return float2(+cossin.x*v.x + cossin.y*v.y,
                -cossin.y*v.x + cossin.x*v.y);
}

float3 inverseRotateXY(in float3 v, in float2 cossin)
{
  return float3(+cossin.x*v.x + cossin.y*v.y,
                -cossin.y*v.x + cossin.x*v.y,
                v.z);
}

float3 waveDisplace(in float2 world, in float lod)
{
  if (lod < lodDisplace.x) { return float3(0, 0, 0); }

  float lodBlend = 1;// saturate(lodDisplace.y*(lod - lodDisplace.x));

  int2 baseTile;
  float2 basePos;
  baseTileCoords(baseTile, basePos, world);

  float4 p00, p01, p10, p11;
  setupTileBlend(p00, p01, p10, p11, baseTile, basePos);

  float3 s00 = decodeDisplacement(Displacement.SampleLevel(DisplacementSampler, p00.xy, 0).rgb);
  float3 s01 = decodeDisplacement(Displacement.SampleLevel(DisplacementSampler, p01.xy, 0).rgb);
  float3 s10 = decodeDisplacement(Displacement.SampleLevel(DisplacementSampler, p10.xy, 0).rgb);
  float3 s11 = decodeDisplacement(Displacement.SampleLevel(DisplacementSampler, p11.xy, 0).rgb);

  // Bilinear weights
  float b00 = (1.0 - basePos.x)*(1.0 - basePos.y);
  float b01 = (1.0 - basePos.x)*basePos.y;
  float b10 = basePos.x*(1.0 - basePos.y);
  float b11 = basePos.x*basePos.y;;


  float3 s = (b00 * inverseRotateXY(s00, p00.zw) +
              b01 * inverseRotateXY(s01, p01.zw) +
              b10 * inverseRotateXY(s10, p10.zw) +
              b11 * inverseRotateXY(s11, p11.zw));

  return lodBlend * significantWaveHeight * s;
}

float3 waveNormal(out float foam, in float2 world, in float lod, in float foamFactor)
{
  if (lod < lodNormal.x) {
    foam = 0.0;
    return float3(0, 0, 1);
  }
  float lodNormalBlend = saturate(lodNormal.y*(lod - lodNormal.x));

  int2 baseTile;
  float2 basePos;
  baseTileCoords(baseTile, basePos, world);

  float4 p00, p01, p10, p11;
  setupTileBlend(p00, p01, p10, p11, baseTile, basePos);
  // Swap tiles such that a tile is always sampled from the same sampler,
  // otherwise the hardware LOD-calculation fails at seams.
  if (baseTile.x & 1) {
    float4 t;
    t = p00; p00 = p10; p10 = t;
    t = p01; p01 = p11; p11 = t;
    basePos.x = 1.f - basePos.x;
  }
  if (baseTile.y & 1) {
    float4 t;
    t = p00; p00 = p01; p01 = t;
    t = p10; p10 = p11; p11 = t;
    basePos.y = 1.f - basePos.y;
  }

  float4 n00 = decodeDifference(Difference.Sample(DifferenceSampler, p00.xy));
  float4 n01 = decodeDifference(Difference.Sample(DifferenceSampler, p01.xy));
  float4 n10 = decodeDifference(Difference.Sample(DifferenceSampler, p10.xy));
  float4 n11 = decodeDifference(Difference.Sample(DifferenceSampler, p11.xy));
  float b00 = (1.0 - basePos.x)*(1.0 - basePos.y);
  float b01 = (1.0 - basePos.x)*basePos.y;
  float b10 = basePos.x*(1.0 - basePos.y);
  float b11 = basePos.x*basePos.y;;

  float3 n =
    inverseRotateXY(n00.xyz, p00.zw)*b00 +
    inverseRotateXY(n01.xyz, p01.zw)*b01 +
    inverseRotateXY(n10.xyz, p10.zw)*b10 +
    inverseRotateXY(n11.xyz, p11.zw)*b11;

  foam = 
    n00.w * b00 +
    n01.w * b01 +
    n10.w * b10 +
    n11.w * b11;

  float detailBlend = 0.0;
  if (lodDetail.x < lod) {
    detailBlend = saturate(lodDetail.y*(lod - lodDetail.x));

    float4 nm = decodeDifference(Difference.Sample(DifferenceSampler, 3.3*fftTileScale*world.xy));
    n += 0.5*detailBlend*nm.xyz;
    foam += 0.5*nm.w;

    float4 nd = decodeDifference(Difference.Sample(DifferenceSampler, 17.1*fftTileScale*world.xy));
    n += 0.25*detailBlend*nd.xyz;

    foam += 0.25*nd.w;
  }
  n = normalize(n);

  foam = 0.0;

  return normalize(lerp(float3(0, 0, 1), n, lodNormalBlend));
}
