#ifndef COGS_INCLUDE_PHONGLIGHTING
#define COGS_INCLUDE_PHONGLIGHTING

#ifndef COGS_MAX_LIGHTS
#define COGS_MAX_LIGHTS 8
#endif

uniform vec4 lightTransforms[COGS_MAX_LIGHTS];      // direction with w=0 if directional, otherwise position with w=radius
uniform vec4 lightColors[COGS_MAX_LIGHTS];          // light color.rgb, w=shadow intensity
uniform ivec4 numLights;
uniform vec4 eyePosition;
uniform vec4 fogColor;        // [fogColor.rgb, fogDistance]
uniform vec4 ambientColor;    // [ambient.rgb, fogAmount]

#ifdef COGS_FORWARD_SHADOWS_ENABLED
#include "Engine/CascadedShadowMaps.es20.glsl"
#include "Engine/PointLightShadowMaps.es20.glsl"
#endif

vec3 PhongLighting(in vec4 lightTransform,
                   in vec4 lightColor,
                   vec3 diffuseColor, vec3 normal, vec3 worldPosition, vec3 specularColor, float specularPower)
{
  float attenuation;
  vec3 lightDir;
  if (lightTransform.w == 0.0) {      // Directional light
    lightDir = lightTransform.xyz;
    attenuation = 1.0;
  }
  else {                              // Point light
    lightDir = worldPosition - lightTransform.xyz;
    float distance = length(lightDir);
    lightDir *= 1.0 / distance;

    float radius = lightTransform.w;
    attenuation = clamp(1.0 - (distance * distance) / (radius * radius), 0.0, 1.0);
    attenuation *= attenuation;
  }

  float diffuseIntensity = clamp(dot(normal, -lightDir), 0.0, 1.0);

  vec3 color = diffuseIntensity * attenuation * diffuseColor.rgb * lightColor.rgb ;

  if (max(max(specularColor.r, specularColor.g), specularColor.b) > 0.0) {
    vec3 eyeDir = -normalize(eyePosition.xyz - worldPosition);
    vec3 halfDir = normalize(lightDir + eyeDir);
    float specularIntensity = pow(clamp(dot(normal, -halfDir), 0.0, 1.0), specularPower);

    color += specularIntensity * attenuation * specularColor * lightColor.rgb;
  }

  return color;
}

struct SurfaceLighting
{
  vec4 Color;
  vec3 Normal;
  vec3 Specular;
  vec3 ViewSpacePosition;
  float SpecularPower;
  vec3 Emissive;
  int LightModel;
  vec4 WorldPosition;
};

struct VolumeLighting
{
  vec4 Color;
  vec3 Normal;
  vec3 Specular;
  vec3 ViewSpacePosition;
  float SpecularPower;
  vec3 Emissive;
  int LightModel;
  vec4 WorldPositionA;
  vec4 WorldPositionB;
};

// Applies simple fog color.
vec3 applySimpleFog(in SurfaceLighting surface, vec3 inputColor)
{
  float distance = length(surface.ViewSpacePosition);

  vec3 viewDir = -normalize(eyePosition.xyz - surface.WorldPosition.xyz);

  vec3 fogCol = fogColor.rgb;

  float fogDensity = 1.0 / fogColor.a;
  float fogAmount = 1.0 - exp(-distance * fogDensity);
  float t = clamp(fogAmount, 0.0, 1.0);

  return mix(inputColor, fogCol, t);
}

vec3 computeFog(in SurfaceLighting surface, vec3 inputColor)
{
  return applySimpleFog(surface, inputColor);
}


vec4 DoLighting(in VolumeLighting volume, vec4 screenPos)
{
  vec4 outColor = vec4(0, 0, 0, 0);

  if (true) { // (volume.LightModel & COGS_LIGHT_MASK_PHONG) == COGS_LIGHT_MASK_PHONG
    vec3 color = vec3(0, 0, 0);

    for (int i = 0; i < COGS_MAX_LIGHTS; ++i) {
      float shadowFactor = 1.0;

#ifdef COGS_FORWARD_SHADOWS_ENABLED
      if (lightTransforms[i].w == 0.0) {
#if COGS_NUM_DIRECTIONAL_SHADOWS > 0
        shadowFactor = 0.0;
        //shadowFactor = calculateShadowContribution(-volume.ViewSpacePosition.z,
        //                                           volume.WorldPositionA,
        //                                           volume.WorldPositionB,
        //                                           lightDirections[i].xyz,
        //                                           i,
        //                                           screenPos);
#endif
      }
#endif

      float shadowIntensityOffset = lightColors[i].w;
      shadowFactor = min(shadowFactor + shadowIntensityOffset, 1.0);

      color += shadowFactor * PhongLighting(lightTransforms[i],
                                            lightColors[i],
                                            volume.Color.rgb, volume.Normal, volume.WorldPositionA.xyz, volume.Specular, volume.SpecularPower);
    }

    color += volume.Color.rgb * ambientColor.rgb;

    outColor = vec4(color + volume.Emissive.rgb, volume.Color.a);
  } else {
    outColor = vec4(volume.Color.rgb + volume.Emissive, volume.Color.a);
  }

  return outColor;
}


vec4 DoLighting(in SurfaceLighting surface, vec4 screenPos, vec4 indirectLighting)
{
  vec4 outColor = vec4(0, 0, 0, 0);

#if defined(COGS_LIGHT_MODEL_PHONG)
  if (true) { // surface.LightModel & COGS_LIGHT_MASK_PHONG) == COGS_LIGHT_MASK_PHONG
    vec3 color = surface.Color.rgb * indirectLighting.rgb;

    for (int i = 0; i < COGS_MAX_LIGHTS; ++i) {
      float shadowFactor = indirectLighting.a;

#ifdef COGS_FORWARD_SHADOWS_ENABLED
      if (lightTransforms[i].w == 0.0) {
#if COGS_NUM_DIRECTIONAL_SHADOWS > 0
        shadowFactor = indirectLighting.a*calculateShadowContribution(surface.WorldPosition, surface.Normal, lightTransforms[i].xyz, i, screenPos);
#endif
      } else {
#if COGS_NUM_POINT_SHADOWS > 0
        shadowFactor = indirectLighting.a * calculatePointLightShadow(lightTransforms[i], surface.WorldPosition.xyz, surface.Normal);
#endif
      }
#endif

      float shadowIntensityOffset = lightColors[i].w;
      shadowFactor = min(shadowFactor + shadowIntensityOffset, 1.0);

#ifdef COGS_IMAGE_BASED_LIGHTING
      if(i==0) {
          vec3 eyeDir = -normalize(eyePosition.xyz - surface.WorldPosition.xyz);
          vec3 refDir = reflect(eyeDir, surface.Normal);

          float roughness = (1.0/128.0)*(150.0 - surface.SpecularPower);
          vec3 spec = surface.Specular.rgb
                    * textureCubeLodEXT(environmentRadiance, refDir.xzy, roughness * environmentRadianceMips()).rgb;

          vec3 diff = surface.Color.rgb
                    * mix(ambientColor.rgb * textureCube(ambientIrradiance, surface.Normal.xzy).rgb,
                          textureCube(environmentIrradiance, surface.Normal.xzy).rgb,
                          shadowFactor);

          color += environmentBrightness() * lightColors[0].rgb * (spec + diff);
          continue;
      }
#endif

      color += shadowFactor * PhongLighting(lightTransforms[i],
                                            lightColors[i],
                                            surface.Color.rgb, surface.Normal, surface.WorldPosition.xyz, surface.Specular, surface.SpecularPower);
    }

#ifndef COGS_IMAGE_BASED_LIGHTING
    color += surface.Color.rgb * ambientColor.rgb;
#endif

    outColor = vec4(color + surface.Emissive.rgb, surface.Color.a);
  }
#elif defined(COGS_LIGHT_MODEL_BASE_COLOR)
  if (true) { // (surface.LightModel & 15) == 0
    outColor = vec4(surface.Color.rgb + surface.Emissive, surface.Color.a);
  }
#endif

#if defined(COGS_INDIRECT_LIGHTING_FULL)
  if (true) { // (surface.LightModel & COGS_LIGHT_MASK_INDIRECT) == COGS_LIGHT_MASK_INDIRECT
    float m = 0.5*max(max(surface.Color.r, surface.Color.g), surface.Color.b);

    outColor += vec4(indirectLighting.rgb * (vec3(m) + surface.Color.rgb) + indirectLighting.a*(surface.Color.rgb + surface.Emissive), surface.Color.a);
  }
#endif 

  if (ambientColor.a != 0.0) {
    outColor.rgb = mix(outColor.rgb, computeFog(surface, outColor.rgb), ambientColor.a);
  }

  return outColor;
}

vec4 DoLighting(in SurfaceLighting surface, vec4 screenPos)
{
  return DoLighting(surface, screenPos, vec4(0, 0, 0, 1));
}


#endif
