#ifndef COGS_INCLUDE_PHONGLIGHTING
#define COGS_INCLUDE_PHONGLIGHTING

#ifndef COGS_MAX_LIGHTS
#define COGS_MAX_LIGHTS 8
#endif


#ifdef COGS_FORWARD_SHADOWS_ENABLED
#include "Engine/CascadedShadowMaps.es30.glsl"
#include "Engine/PointLightShadowMaps.es30.glsl"
#endif

vec3 PhongLighting(int i, vec3 diffuseColor, vec3 normal, vec3 worldPosition, vec3 specularColor, float specularPower)
{
  vec4 lightPosition = lightPositions[i];

  vec3 offset = worldPosition - lightPosition.xyz;
  vec3 lightDir = lightDirections[i].w == 0.f ? lightDirections[i].xyz : normalize(offset);

  float distance = length(offset);
  float radius = lightParameters[i].a;

  float attenuation = lightDirections[i].w == 0.0 ? 1.0 : clamp(1.0 - (distance * distance) / (radius * radius), 0.0, 1.0);
  attenuation *= attenuation;

  float lightIntensity = attenuation * lightColorIntensity[i].a;
  float diffuseIntensity = clamp(dot(normal, -lightDir), 0.0, 1.0);

  vec3 color = diffuseColor.rgb * lightColorIntensity[i].rgb * diffuseIntensity * lightIntensity;

  if (max(max(specularColor.r, specularColor.g), specularColor.b) > 0.0) {
    vec3 eyeDir = -normalize(eyePosition.xyz - worldPosition);
    vec3 halfDir = normalize(lightDir + eyeDir);
    float specularIntensity = pow(clamp(dot(normal, -halfDir), 0.0, 1.0), specularPower);

    color += specularColor * lightColorIntensity[i].rgb * specularIntensity * lightIntensity;
  }
  return color;
}

struct SurfaceLighting
{
  vec4 Color;
  vec3 Normal;
  vec3 Specular;
  vec3 ViewSpacePosition;
  float SpecularPower;
  vec3 Emissive;
  int LightModel;
  vec4 WorldPosition;
};

struct VolumeLighting
{
  vec4 Color;
  vec3 Normal;
  vec3 Specular;
  vec3 ViewSpacePosition;
  float SpecularPower;
  vec3 Emissive;
  int LightModel;
  vec4 WorldPositionA;
  vec4 WorldPositionB;
};

// Applies simple fog color.
vec3 applySimpleFog(in SurfaceLighting surface, vec3 inputColor)
{
  float distance = length(surface.ViewSpacePosition);

  vec3 viewDir = -normalize(eyePosition.xyz - surface.WorldPosition.xyz);

#ifdef COGS_IMAGE_BASED_LIGHTING
  vec3 fogCol = fogColor.rgb * environmentBrightness * textureLod(environmentRadiance, (1.0 / distance) * viewDir.xzy, 3.0).rgb;
#else
  vec3 fogCol = fogColor.rgb;
#endif
  if ((fogEnabled & 3) == 2) {

    fogCol = environmentBrightness * textureLod(environmentIrradiance,  (1.0 / distance) * viewDir.xzy, 0.0).rgb;
  }

  float t;
  if ((fogEnabled & 4) != 0) {
    // Total occulsion behaviour
    // 1 - e^(-5) is pretty flat and close to 1. We make it so that
    // the function is exactly one here so we can cull geometry
    // further away. Useful for dense fog.
    // 1/(1-e^{-5})=1.0068.
    float fogDensity = 5.0 / fogDistance;
    t = min(1.0, 1.0068*(1.0 - exp(-distance * fogDensity)));
  } else {
    // classic behaviour
    const float maxFogAmount = 1.0;
    float fogDensity = 1.0 / fogDistance;
    float fogAmount = 1.0 - exp(-distance * fogDensity);
    t = clamp(fogAmount, 0.0, maxFogAmount);
  }

  return mix(inputColor, fogCol, t);
}

vec3 computeFog(in SurfaceLighting surface, vec3 inputColor)
{
  return applySimpleFog(surface, inputColor);
}


vec4 DoLighting(in VolumeLighting volume, vec4 screenPos)
{
  vec4 outColor = vec4(0, 0, 0, 0);

  if ((volume.LightModel & COGS_LIGHT_MASK_PHONG) == COGS_LIGHT_MASK_PHONG) {
    vec3 color = vec3(0, 0, 0);

    for (int i = 0; i < numLights.x; ++i) {
      float shadowFactor = 1.0;

#ifdef COGS_FORWARD_SHADOWS_ENABLED
      if ( ((volume.LightModel & COGS_LIGHT_MASK_SHADOW_RECEIVER) == COGS_LIGHT_MASK_SHADOW_RECEIVER) && (lightPositions[i].w == 0.0)) {
#if COGS_NUM_DIRECTIONAL_SHADOWS > 0
        shadowFactor = calculateShadowContribution(-volume.ViewSpacePosition.z,
                                                   volume.WorldPositionA,
                                                   volume.WorldPositionB,
                                                   lightDirections[i].xyz,
                                                   i,
                                                   screenPos);
#endif
      }
#endif

      float shadowIntensityOffset = lightParameters[i].r;
      shadowFactor = min(shadowFactor + shadowIntensityOffset, 1.0);

      color += shadowFactor * PhongLighting(i, volume.Color.rgb, volume.Normal, volume.WorldPositionA.xyz, volume.Specular, volume.SpecularPower);
    }

    color += volume.Color.rgb * ambientColor * ambientIntensity;

    outColor = vec4(color + volume.Emissive.rgb, volume.Color.a);
  } else {
    outColor = vec4(volume.Color.rgb + volume.Emissive, volume.Color.a);
  }

  return outColor;
}


vec4 DoLighting(in SurfaceLighting surface, vec4 screenPos, vec4 indirectLighting)
{
  vec4 outColor = vec4(0, 0, 0, 0);

  if ((surface.LightModel & COGS_LIGHT_MASK_PHONG) == COGS_LIGHT_MASK_PHONG) {
    vec3 color = surface.Color.rgb * indirectLighting.rgb;
    for (int i = 0; i < numLights.x; ++i) {
      vec4 lightPosition = lightPositions[i];
      float shadowFactor = indirectLighting.a;

#ifdef COGS_FORWARD_SHADOWS_ENABLED
      if (((surface.LightModel & COGS_LIGHT_MASK_SHADOW_RECEIVER) == COGS_LIGHT_MASK_SHADOW_RECEIVER) && (lightPositions[i].w == 0.0)) {
#if COGS_NUM_DIRECTIONAL_SHADOWS > 0
        shadowFactor = indirectLighting.a * calculateShadowContribution(surface.WorldPosition, surface.Normal, lightDirections[i].xyz, i, screenPos);
#endif
      } else {
#if COGS_NUM_POINT_SHADOWS > 0
        shadowFactor = indirectLighting.a * calculatePointLightShadow(lightPosition, surface.WorldPosition.xyz, surface.Normal, i);
#endif
      }
#endif

      float shadowIntensityOffset = lightParameters[i].r;
      shadowFactor = min(shadowFactor + shadowIntensityOffset, 1.0);

#ifdef COGS_IMAGE_BASED_LIGHTING
      if (i == 0) {
        vec3 eyeDir = -normalize(eyePosition.xyz - surface.WorldPosition.xyz);
        vec3 refDir = reflect(eyeDir, surface.Normal);

        float l = 0.001*max(0.0, 80.0 - surface.SpecularPower) * environmentRadianceMips();

        color += shadowFactor * environmentBrightness * (
          surface.Color.rgb * textureLod(environmentIrradiance, surface.Normal.xzy, 0.0).rgb +
          surface.Specular.rgb * textureLod(environmentRadiance, refDir.xzy, l).rgb);

        continue;
      }
#endif
      color += shadowFactor * PhongLighting(i, surface.Color.rgb, surface.Normal, surface.WorldPosition.xyz, surface.Specular, surface.SpecularPower);
   
    }

#ifdef COGS_IMAGE_BASED_LIGHTING
    color += surface.Color.rgb * ambientIntensity * textureLod(ambientIrradiance, surface.Normal.xzy, 0.0).rgb;
#else
    color += surface.Color.rgb * ambientColor * ambientIntensity;
#endif

    outColor = vec4(color + surface.Emissive.rgb, surface.Color.a);
  } else if ((surface.LightModel & 15) == 0) {
    outColor = vec4(surface.Color.rgb + surface.Emissive, surface.Color.a);
  }
  
  if ((surface.LightModel & COGS_LIGHT_MASK_INDIRECT) == COGS_LIGHT_MASK_INDIRECT) {
    float m = 0.5*max(max(surface.Color.r, surface.Color.g), surface.Color.b);

    outColor += vec4(indirectLighting.rgb * vec3(m + surface.Color.rgb) + indirectLighting.a*(surface.Color.rgb + surface.Emissive), surface.Color.a);
  }
  
  if ((fogEnabled != 0) && (fogAmount != 0.0)) {
    outColor.rgb = mix(outColor.rgb, computeFog(surface, outColor.rgb), fogAmount);
  }

  return outColor;
}

vec4 DoLighting(in SurfaceLighting surface, vec4 screenPos)
{
  return DoLighting(surface, screenPos, vec4(0, 0, 0, 1));
}

#endif
