#include "Engine/TransparencyBlendCommon.hlsl"
#include "Engine/PhongLighting.hlsl"

Texture2D depthTexture;
SamplerState linearSampler;

ByteAddressBuffer               List;
StructuredBuffer<FragmentNode>  Node;
StructuredBuffer<FragmentData>  Data;

struct SurfaceSample
{
  float4 Color;
  float4 Emissive;
  float3 Normal;
  float Depth;
  float4 ViewSpacePosition;
  float Facing;
};

PixelOut main(VertexIn In, float4 ScreenPos : SV_Position)
{
  const float backgroundDepth = depthTexture.Load(int3(ScreenPos.xy, 0)).r;

  uint startOffset = 4 * ((uint)ScreenPos.y/HeightDivisor * Width + (uint)ScreenPos.x/WidthDivisor);
  uint firstNodeIndex = List.Load(startOffset);

  if (NodeAlloc <= (firstNodeIndex & 0x7fffffffu)) {
    PixelOut BackOut = { float4(0, 0, 0, 1) };
    return BackOut;
  }

  float doneDepth = -1e6;
  uint doneIndex = 0;

  SurfaceSample surfaceSamples[20];
  uint numSamples = 0;

  for (int q = 0; q < 20; q++) {
    // Find the nearest fragment that is further away than ProcessedDepth.
    float nextDepth = backgroundDepth;
    uint nextIndex = firstNodeIndex;

    uint maxIter = 100;
    uint currentIter = 0;
    uint index = firstNodeIndex;
    do {
      if (currentIter++ > maxIter) break;

      float depth = Node[index & 0x7fffffffu].Depth;

#ifdef COGS_REVERSE_DEPTH
      bool doneTest = (doneDepth > depth) || ((doneDepth == depth) && (doneIndex < index));
      bool nextTest = (depth > nextDepth) || ((depth == nextDepth) && (index < nextIndex));
#else
      bool doneTest = (doneDepth < depth) || ((doneDepth == depth) && (doneIndex < index));
      bool nextTest = (depth < nextDepth) || ((depth == nextDepth) && (index < nextIndex));
#endif

      if (doneTest && nextTest) {
        nextDepth = depth;
        nextIndex = index;
      }
      index = Node[index & 0x7fffffffu].Next;
    } while ((index & 0x7fffffffu) < NodeAlloc);

    // We have processed the entire list.
#ifdef COGS_REVERSE_DEPTH
   if (backgroundDepth >= nextDepth) break;
#else
   if (backgroundDepth <= nextDepth) break;
#endif

    // Fetch fragment
    float depth = Node[nextIndex & 0x7fffffffu].Depth;
    FragmentData data = Data[nextIndex & 0x7fffffffu];

    // Process sample.
    float4 viewSpacePosition = mul(float4(In.ScreenPosition.xy, depth, 1), inverseProjectionMatrix);
    viewSpacePosition /= viewSpacePosition.w;

    SurfaceSample surfaceSample;

    int LightModel;
    float4 Specular;
    unpack(surfaceSample.Normal,
           LightModel,
           surfaceSample.Color.a,
           surfaceSample.Emissive.rgb,
           surfaceSample.Color.rgb,
           Specular.rgb,
           Specular.a,
           data);

    surfaceSample.Depth = depth;
    surfaceSample.ViewSpacePosition = viewSpacePosition;
    surfaceSample.Facing = (nextIndex & 0x80000000u) != 0 ? 1.0 : -1.0;
    
    surfaceSamples[q] = surfaceSample;
    numSamples++;

    doneDepth = nextDepth;
    doneIndex = nextIndex;
  }

  float4 result = float4(0, 0, 0, 1);
  for (uint i = 0; i < numSamples; ++i) {
    SurfaceSample s = surfaceSamples[numSamples - i - 1];

    SurfaceLighting lightSurface = (SurfaceLighting)0;
    lightSurface.Color = s.Color;
    lightSurface.Emissive = s.Emissive.rgb;
    lightSurface.Normal = s.Normal;
    lightSurface.ViewSpacePosition = s.ViewSpacePosition.xyz;
    lightSurface.SpecularPower = 80;
    lightSurface.Specular = float3(1, 1, 1);
    lightSurface.LightModel = 1;

    float3 l = surfaceSamples[numSamples - i].ViewSpacePosition.xyz - s.ViewSpacePosition.xyz;
    float3x3 inverseViewRot = inverseViewMatrix;
    l = mul(l, inverseViewRot);

    float distance = abs(length(l));

    float C = 0.65;

    float3 absorptionColor = float3(0.8, 0.8, 0.8);

    float minAbs = 0.2;
    float maxAbs = 0.9;

    float3 absorbance = -distance * C * absorptionColor;
    float3 transparency = clamp(exp(absorbance), minAbs.rrr, maxAbs.rrr);
    float3 alpha = 1 - transparency;

    float4 shaded = float4(DoLighting(lightSurface, ScreenPos).rgb, 1);

    if (s.Facing) {
      result.rgb = result.rgb*transparency + shaded.rgb*alpha;
      result.a = result.a*transparency;
    }
  }

  PixelOut Out;
  Out.Color = result;
  return Out;
}
