#include "Engine/TransparencyBlendCommon.hlsl"
#include "Engine/PhongLighting.hlsl"

Texture2D depthTexture;
SamplerState linearSampler;

ByteAddressBuffer               List;
StructuredBuffer<FragmentNode>  Node;
StructuredBuffer<FragmentData>  Data;

PixelOut main(VertexIn In, float4 ScreenPos : SV_Position)
{
  const float backgroundDepth = depthTexture.Load(int3(ScreenPos.xy, 0)).r;

  uint startOffset = 4 * ((uint)ScreenPos.y/HeightDivisor * Width + (uint)ScreenPos.x/WidthDivisor);
  uint rx = List.Load(startOffset);
  uint ix = rx & 0x7fffffffu;

  if (NodeAlloc <= ix) {
    PixelOut BackOut = { float4(0, 0, 0 ,1) };
    return BackOut;
  }

  // Plow through the full fragment list for a pixel location _once_, while maintaining
  // a depth-sorted array of the MAX_SAMPLES nearest entries.
  uint indices[MAX_SAMPLES];
  float depths[MAX_SAMPLES];
  int tailIx = 0;
  indices[0] = rx;
  depths[0] = Node[ix].Depth;
  rx = Node[ix].Next;
  ix = rx & 0x7fffffffu;

  uint maxIter = 100;
  uint currentIter = 0;
  while (ix < NodeAlloc) {
    // Driver crash safeguard.
    if (currentIter++ == maxIter) break;

    float dp = Node[ix].Depth;
    bool notFull = tailIx < MAX_SAMPLES - 1;
#ifdef COGS_REVERSE_DEPTH
    bool closerThanTail = dp > depths[tailIx];
#else
    bool closerThanTail = dp < depths[tailIx];
#endif
    if (notFull || closerThanTail) {
      tailIx = tailIx + (notFull ? 1 : 0);
      int j = tailIx;
      for (int i = tailIx - 1; i >= 0; i--) {
#ifdef COGS_REVERSE_DEPTH
        if (dp > depths[i]) {
#else
        if (dp < depths[i]) {
#endif
          indices[i + 1] = indices[i];
          depths[i + 1] = depths[i];
          j = i;
        }
      }
      indices[j] = rx;
      depths[j] = dp;
    }
    rx = Node[ix].Next;
    ix = rx & 0x7fffffffu;
  }

  float maxAlpha = 0.0;
  float4 normalBlend = float4(0, 0, 0, 0);
  float4 maxBlend = float4(0, 0, 0, 0);
  for (int i = 0; (i <= tailIx) && (max(normalBlend.a,maxBlend.a) < 0.99); i++) {
    uint ix = indices[i];
    float depth = depths[i];
    FragmentData data = Data[ix & 0x7fffffffu];
    int LightModel = data.Emissive >> 24;

    bool closed = (LightModel & 16) != 0;
    bool entering = (ix & 0x80000000u) != 0;
    if (closed && !entering) {
      // Skip back-sides of closed objects.
      continue;
    }

    float4 viewSpacePosition = mul(float4(In.ScreenPosition.xy, depth, 1), inverseProjectionMatrix);
    viewSpacePosition /= viewSpacePosition.w;
    float4 worldSpacePosition = mul(viewSpacePosition, inverseViewMatrix);
    SurfaceLighting lightSurface = (SurfaceLighting)0;
    lightSurface.ViewSpacePosition = viewSpacePosition.xyz;
    unpack(lightSurface.Normal,
           lightSurface.LightModel,
           lightSurface.Color.a,
           lightSurface.Emissive,
           lightSurface.Color.rgb,
           lightSurface.Specular,
           lightSurface.SpecularPower,
           data);
    float4 shaded = float4(DoLighting(lightSurface, ScreenPos).rgb, lightSurface.Color.a);
  
    if (closed) {
      if (maxAlpha < shaded.a) {
        maxAlpha = shaded.a;
        maxBlend = shaded;
      }
      else {
        maxBlend = blend(maxBlend, shaded);
      }
    }
    else {

      if (maxAlpha != 0) {
        normalBlend = blend(normalBlend, maxBlend);
        maxBlend = float4(0, 0, 0, 0);
        maxAlpha = 0;
      }
      normalBlend = blend(normalBlend, shaded);
    }
  }

  if (maxAlpha != 0) {
    normalBlend = blend(normalBlend, maxBlend);
  }

  if (normalBlend.a >= 0.999)
    normalBlend.a = 1.0;

  PixelOut Out;
  Out.Color.rgb = normalBlend.rgb*normalBlend.a;
  Out.Color.a = 1.0-normalBlend.a;
  return Out;
}
