precision mediump float;

uniform highp mat4 viewMatrix;

uniform sampler2D diffuseTexture;

varying vec4 vPosition;
varying vec3 vNormal;
varying vec4 vColor;
varying vec2 vTexCoord;

struct Light
{
  vec4 position;
  vec4 diffuseColor;
  vec4 specularColor;
  vec4 ambientColor;
};

uniform Light lights[MAX_LIGHTS];
uniform bool lightEnabled[MAX_LIGHTS];

uniform vec4 diffuseColor;
uniform vec4 specularColor;
uniform vec4 ambientColor;
uniform vec4 emissiveColor;

uniform vec4 fogColor;
uniform float fogDistance;
uniform bool fogEnabled;

vec3 applyFog(vec3 color, vec3 position)
{
  vec3 viewPosition = (viewMatrix * vec4(position, 1)).xyz;
  float distance = length(viewPosition);

  float fogDensity = 1.0 / fogDistance;
  float fogAmount = 1.0 - exp(-distance * fogDensity);

  vec3 foggedColor = mix(color, fogColor.rgb, clamp(fogAmount, 0.0, 1.0));

  return mix(color, foggedColor, fogEnabled ? 1.0 : 0.0);
}

vec4 calculateLights(vec3 normal, vec3 position, Light light)
{
  vec3 lightDir = light.position.w == 0.0 ? (viewMatrix * light.position).xyz : normalize(position - (viewMatrix * vec4(light.position.xyz, 1)).xyz);

  vec3 eyeDir = vec3(0, 0, -1);
  vec3 halfDir = normalize(lightDir + eyeDir);

  float specularPower = 100.0;

  float specular = pow(clamp(dot(normal, -halfDir), 0.0, 1.0), specularPower);
  float intensity = clamp(dot(normal, -lightDir), 0.0, 1.0);

  vec4 diffuseLight = intensity * light.diffuseColor * diffuseColor;
  vec4 specularLight = specular * light.specularColor * specularColor;
  vec4 ambientLight = light.ambientColor * ambientColor;

  return diffuseLight + specularLight + ambientLight + emissiveColor;
}

void main()
{
  vec3 normal = normalize(vNormal);
  vec3 position = vPosition.xyz;

#ifdef ENABLE_LIGHTING
  vec4 lightColor = vec4(0, 0, 0, 0);

  for (int i = 0; i < MAX_LIGHTS; i++) {
    if (lightEnabled[i]) {
      if (gl_FrontFacing) {
        lightColor += calculateLights(normal, position, lights[i]);
      } else {
        lightColor += calculateLights(-normal, position, lights[i]);
      }
    }
  }

  lightColor.a = diffuseColor.a;
#else
  vec4 lightColor = diffuseColor + emissiveColor;
#endif

#ifdef ENABLE_TEXTURE
  vec4 textureColor = texture2D(diffuseTexture, vTexCoord);
#else
  vec4 textureColor = vec4(1, 1, 1, 1);
#endif

  vec4 color = vColor * textureColor * lightColor;

  gl_FragColor = vec4(applyFog(color.rgb, position), color.a);
}
