#include "Engine/Common.hlsl"

Texture2D depth4xTexture;
Texture2D normalsTexture;
SamplerState pointClampSampler;

struct VertexIn
{
  float4 Position : SV_Position;
  float4 ScreenPosition : TEXCOORD0;
  float2 TexCoords : TEXCOORD1;
};

cbuffer BasicAmbientOcclusionParameters
{
  float worldSpaceRadius;
  float maxScreenRadius;
  float contrast;
};

#define VARIANT 1
#define PI 3.1415f
#define TWO_PI 2.0f * PI
#define GOLDEN_ANGLE 2.4f

#define SAMPLES_COUNT 16

static float2 vogelDiskOffsets16[16] =
{
  float2(0.176777f, 0.0f),
  float2(-0.225718f, 0.206885f),
  float2(0.0343507f, -0.393789f),
  float2(0.284864f, 0.370948f),
  float2(-0.52232f, -0.0918239f),
  float2(0.494281f, -0.315336f),
  float2(-0.164493f, 0.615786f),
  float2(-0.316681f, -0.607012f),
  float2(0.685167f, 0.248588f),
  float2(-0.711557f, 0.295696f),
  float2(0.341422f, -0.73463f),
  float2(0.256072f, 0.808194f),
  float2(-0.766143f, -0.440767f),
  float2(0.896453f, -0.200303f),
  float2(-0.544632f, 0.780785f),
  float2(-0.130341f, -0.975582f)
};

float2 RotatePoint(float2 pt, float angle)
{
  float sine, cosine;
  sincos(angle, sine, cosine);

  float2 rotatedPoint;
  rotatedPoint.x = cosine*pt.x + -sine*pt.y;
  rotatedPoint.y = sine*pt.x + cosine*pt.y;

  return rotatedPoint;
}

float3 ViewPosition(float2 texCoords)
{
  float2 planeSize = max(float2(1, 1 / projectionParameters.z), projectionParameters.xy);

  float3 position = float3(texCoords, -1.0f);
  position.xy = position.xy - 0.5f;
  position.y *= -1.0f;
  position.xy *= planeSize;
  position *= -depth4xTexture.Sample(pointClampSampler, texCoords).r;

  return position;
}

float2 VogelDiskOffset(int sampleIndex, float phi)
{
  float r = sqrt(sampleIndex + 0.5f) / sqrt(SAMPLES_COUNT);
  float theta = sampleIndex * GOLDEN_ANGLE + phi;

  float sine, cosine;
  sincos(theta, sine, cosine);

  return float2(r * cosine, r * sine);
}

float InterleavedGradientNoise(float2 screenPos)
{
  float3 magic = float3(0.06711056f, 4.0f * 0.00583715f, 52.9829189f);

  return frac(magic.z * frac(dot(screenPos, magic.xy)));
}

float4 main(VertexIn In) : SV_Target
{
  float2 screenSize;
  normalsTexture.GetDimensions(screenSize.x, screenSize.y);

  float3 viewPosition = ViewPosition(In.TexCoords);
  float3 normal = normalsTexture.Sample(pointClampSampler, In.TexCoords).xyz;
  normal = normalize(mul(normal, (float3x3)viewMatrix));

  float noise = InterleavedGradientNoise(In.Position.xy);

  float aspect = screenSize.x / screenSize.y;

  float2 screenRadius = worldSpaceRadius / viewPosition.z;
  screenRadius = min(screenRadius, maxScreenRadius);
  screenRadius.y *= aspect;

  float ao = 0.0f;

  for (int i = 0; i < SAMPLES_COUNT; i++) {
    float2 sampleOffset = 0.0f;

    {
#ifdef AO_ROTATE
      sampleOffset = vogelDiskOffsets16[i];
      sampleOffset = RotatePoint(sampleOffset, TWO_PI * noise);
#else
      sampleOffset = VogelDiskOffset(i, TWO_PI * noise);
#endif
    }

    float2 sampleTexCoord = In.TexCoords + screenRadius * sampleOffset;

    float3 samplePosition = ViewPosition(sampleTexCoord);
    float3 v = samplePosition - viewPosition;

    ao += max(0.0f, dot(v, normal) + 0.004f * viewPosition.z) / (dot(v, v) + 0.001f);
  }

  ao = saturate(ao / SAMPLES_COUNT);
  ao = 1.0f - ao;
  ao = pow(ao, contrast);

  return ao;
}
