#include "Standard/StandardLighting.hlsl"
#include "Standard/StandardInput.hlsl"

SurfaceOut surfaceFunction(VertexIn In)
{
  SurfaceOut Out = (SurfaceOut)0;

  float3 viewDir = normalize(eyePosition.xyz - In.WorldPosition.xyz);

  float3 normal = GetNormal(In);
  float4 albedo = GetAlbedo(In);
  float occlusion = GetOcclusion(In);
  float3 emissive = GetEmissive(In);

#if defined(COGS_STANDARD_MATERIAL_ALBEDO_PER_VERTEX) || defined(COGS_STANDARD_MATERIAL_OUTPUT_ALBEDO_PER_VERTEX)
  albedo *= In.Albedo;
#endif

#ifdef COGS_STANDARD_MATERIAL_METALLIC_WORKFLOW
  float metallic = GetMetallic(In);
  float roughness = GetRoughness(In);

  float F0 = lerp(0.04, 1, metallic);
#ifdef COGS_GLTF_FIXES
  float3 dielectricSpecular = float3(0.04, 0.04, 0.04);
  float3 specularColor = lerp(dielectricSpecular, albedo.rgb, metallic);
  float3 diffuseColor = lerp(albedo.rgb * (1 - dielectricSpecular.r), float3(0, 0, 0), metallic);
#else
  float3 specularColor = F0 * lerp(float3(F0, F0, F0), albedo.rgb, metallic);
  float3 diffuseColor = (1 - F0) * lerp(albedo.rgb, float3(0, 0, 0), metallic);
   #endif
#else
  float roughness = 1.0 - GetGloss(In);
  float3 specular = GetSpecular(In);

  float F0 = max(specular.r, max(specular.g, specular.b));

  float3 specularColor = specular;
  float3 diffuseColor = (1 - F0) * albedo.rgb;
#endif
#ifdef COGS_GLTF_FIXES
  roughness = sqrt(roughness);
#endif

  float4 indirectLighting = float4(0, 0, 0, 1);
  float3 color = float3(0, 0, 0);

  for (uint i = 0; i < COGS_MAX_LIGHTS; ++i) {
    if (i >= numLights.x) break;

    float3 offset = In.WorldPosition.xyz - lightPositions[i].xyz;
    float3 lightDir = lightDirections[i].w == 0.f ? -lightDirections[i].xyz : -normalize(offset);
    float distance = length(offset);
    float range = lightParameters[i].a;

    float attenuation = lightDirections[i].w == 0 ? 1 : saturate(1.0 - (distance * distance) / (range * range));
    attenuation *= attenuation;

    float shadowFactor = 1.0;

#ifdef COGS_FORWARD_SHADOWS_ENABLED
    if (lightDirections[i].w == 0) {
#if COGS_NUM_DIRECTIONAL_SHADOWS > 0
      shadowFactor = indirectLighting.a * calculateShadowContribution(In.WorldPosition, normal, lightDir, i, In.ScreenPos);
#endif
    } else {
#if COGS_NUM_POINT_SHADOWS > 0
      shadowFactor = indirectLighting.a * calculatePointLightShadow(lightPositions[i], In.WorldPosition.xyz, normal, i);
#endif
    }
#endif

    color += shadowFactor * StandardLighting(lightColorIntensity[i].rgb * lightColorIntensity[i].a * attenuation, normal, viewDir, lightDir, specularColor, diffuseColor, roughness, F0, occlusion);
  }

  color += StandardAmbient(normal, viewDir, specularColor, diffuseColor, roughness, F0, occlusion);

  Out.Normal = normal;
  Out.Color = float4(color + emissive, albedo.a);

  return Out;
}
